use std::{
    io::{BufWriter, Write},
    path::Path,
};

use pulldown_cmark::{CowStr, MetadataBlockKind, Options, Parser};

pub fn compile_md_file(
    src: impl AsRef<Path>,
    html: impl AsRef<Path>,
    gmi: impl AsRef<Path>,
) -> std::io::Result<()> {
    let src_text = std::fs::read_to_string(src)?;

    let p = ParserVars {
        html: create_file(html)?,
        gmi: create_file(gmi)?,

        state: State::Start,
        ol: false,
        counter: 0,

        links: vec![],
        notes: vec![],
    };

    render_markdown(&src_text, p)
}

#[inline(always)]
fn create_file(path: impl AsRef<Path>) -> std::io::Result<Box<BufWriter<std::fs::File>>> {
    let f = std::fs::OpenOptions::new()
        .create(true)
        .write(true)
        .open(path)?;
    Ok(Box::new(BufWriter::new(f)))
}

fn render_markdown<'md>(src: &'md str, mut p: ParserVars<'md>) -> std::io::Result<()> {
    for event in Parser::new_ext(&src, Options::all()) {
        use pulldown_cmark::Event::*;

        dbg!(&event);

        match event {
            Start(tag) => {
                use pulldown_cmark::{CodeBlockKind, HeadingLevel, Tag::*};

                match tag {
                    Paragraph => {
                        p.html.write_all(b"<p>")?;
                        write_paragraph_start(&mut p)?;
                    }

                    Heading { level, id, .. } => {
                        if let Some(id) = id {
                            p.html
                                .write_fmt(format_args!("<{} id=\"{}\">", level, id))?;
                        } else {
                            p.html.write_fmt(format_args!("<{}>", level))?;
                        }

                        write_paragraph_start(&mut p)?;
                        let hashes = match level {
                            HeadingLevel::H1 => "# ",
                            HeadingLevel::H2 => "## ",
                            _ => "### ",
                        };
                        p.gmi.write_all(hashes.as_bytes())?;
                        p.state = State::Paragraph;
                    }

                    BlockQuote(_kind) => {
                        p.html.write_all(b"<blockquote>")?;

                        write_paragraph_start(&mut p)?;
                        p.gmi.write_all(b"> ")?;
                        p.state = State::Quote;
                    }

                    CodeBlock(CodeBlockKind::Fenced(lang)) => {
                        // TODO: highlighting with syntect
                        p.html.write_all(b"<pre><code>")?;

                        write_paragraph_start(&mut p)?;
                        p.gmi.write_fmt(format_args!("```{}\r\n", lang))?;
                    }

                    CodeBlock(CodeBlockKind::Indented) => {
                        p.html.write_all(b"<pre><code>")?;

                        write_paragraph_start(&mut p)?;
                        p.gmi.write_all(b"```")?;
                    }

                    List(None) => {
                        p.html.write_all(b"<ul>")?;
                        write_paragraph_start(&mut p)?;
                        p.state = State::Start;
                    }

                    List(Some(counter_start)) => {
                        p.counter = counter_start;
                        p.ol = true;

                        if counter_start == 1 {
                            p.html.write_all(b"<ol>")?;
                        } else {
                            p.html
                                .write_fmt(format_args!("<ol start=\"{}\">", counter_start))?;
                        }

                        write_paragraph_start(&mut p)?;
                        p.state = State::Start;
                    }

                    Item => {
                        p.html.write_all(b"<li>")?;

                        if p.state != State::Start {
                            p.gmi.write_all(b"\r\n")?;
                            p.state = State::Paragraph;
                        }
                        if p.ol {
                            p.gmi.write_fmt(format_args!("{}. ", p.counter))?;
                            p.counter += 1;
                        } else {
                            p.gmi.write_all(b"* ")?;
                        }
                    }

                    Strong => {
                        p.html.write_all(b"<b>")?;
                        write_inline(&mut p, "**")?;
                    }

                    Emphasis => {
                        p.html.write_all(b"<i>")?;
                        write_inline(&mut p, "*")?;
                    }

                    Strikethrough => {
                        p.html.write_all(b"<s>")?;
                        write_inline(&mut p, "~")?;
                    }

                    Link { dest_url, .. } => {
                        p.html
                            .write_fmt(format_args!("<a href=\"{}\">", dest_url))?;
                        p.links.push(GmiLink::from_url(dest_url));
                        p.state = State::Link;
                    }

                    Image { dest_url, .. } => {
                        p.html.write_fmt(format_args!(
                            "<figure><img src=\"{}\" alt=\"Image\"><figcaption>",
                            dest_url
                        ))?;
                        p.links.push(GmiLink::from_url(dest_url));
                        p.state = State::Link;
                    }

                    Table(_align) => {
                        p.html.write_all(b"<table>")?;

                        write_paragraph_start(&mut p)?;
                        // gmi ??
                    }

                    TableHead => {
                        p.html.write_all(b"<thead><tr>")?;
                        p.state = State::TableHead;
                    }

                    TableRow => {
                        p.html.write_all(b"<tr>")?;
                    }

                    TableCell => {
                        if p.state == State::TableHead {
                            p.html.write_all(b"<th>")?;
                        } else {
                            p.html.write_all(b"<td>")?;
                        }
                    }

                    FootnoteDefinition(id) => {
                        p.notes.push(Footnote::from_id(id));
                        p.state = State::Footnote;
                    }

                    MetadataBlock(kind) => {
                        p.state = State::Metadata {
                            was_at_start: p.state == State::Start,
                            kind,
                        };
                    }

                    other => {
                        eprintln!("Unsupported tag: {:?}", other);
                    }
                }
            }

            End(tag) => {
                use pulldown_cmark::TagEnd::*;

                match tag {
                    Paragraph => {
                        p.html.write_all(b"</p>")?;

                        if !p.links.is_empty() {
                            p.gmi.write_all(b"\r\n")?;
                            for (i, link) in p.links.iter().enumerate() {
                                p.gmi.write_fmt(format_args!(
                                    "\r\n=> {} [{}]: {}", // => https://... [1]: example
                                    link.url, i, link.title,
                                ))?;
                            }
                            p.links.clear();
                        }
                    }
                    Heading(level) => p.html.write_fmt(format_args!("</{}>", level))?,
                    BlockQuote(_) => p.html.write_all(b"</blockquote>")?,
                    CodeBlock => {
                        p.html.write_all(b"</code></pre>")?;
                        p.gmi.write_all(b"```")?;
                    }
                    List(ordered) => {
                        if ordered {
                            p.html.write_all(b"</ol>")?;
                            p.ol = false;
                        } else {
                            p.html.write_all(b"</ul>")?;
                        }
                    }
                    Item => p.html.write_all(b"</li>")?,
                    Strong => {
                        p.html.write_all(b"</b>")?;
                        write_inline(&mut p, "**")?;
                    }
                    Emphasis => {
                        p.html.write_all(b"</i>")?;
                        write_inline(&mut p, "*")?;
                    }
                    Strikethrough => {
                        p.html.write_all(b"</s>")?;
                        write_inline(&mut p, "~")?;
                    }
                    Link => {
                        p.html.write_all(b"</a>")?;
                        p.gmi.write_fmt(format_args!("[{}]", p.links.len()))?;
                    }
                    Image => {
                        p.html.write_all(b"</figcaption></figure>")?;
                        p.gmi.write_fmt(format_args!("[{}]", p.links.len()))?;
                    }
                    TableHead => p.html.write_all(b"</tr></thead><tbody>")?,
                    TableRow => p.html.write_all(b"</tr>")?,
                    TableCell => {
                        if p.state == State::TableHead {
                            p.html.write_all(b"</th>")?
                        } else {
                            p.html.write_all(b"</td>")?
                        }
                    }
                    Table => p.html.write_all(b"</tbody></table>")?,
                    _ => {}
                }

                p.state = match p.state {
                    State::Metadata {
                        was_at_start: true, ..
                    } => State::Start,
                    _ => State::Paragraph,
                };
            }

            Text(text) => {
                match p.state {
                    State::Metadata { .. } => {
                        // TODO: parse yaml/toml
                        continue;
                    }
                    _ => {}
                }

                p.html.write_all(text.as_bytes())?;
                write_inline(&mut p, &text)?;
            }

            Code(code) => {
                p.html.write_all(b"<code>")?;
                p.html.write_all(code.as_bytes())?;
                p.html.write_all(b"</code>")?;

                write_inline(&mut p, "`")?;
                write_inline(&mut p, &code)?;
                write_inline(&mut p, "`")?;
            }

            SoftBreak => {
                p.html.write_all(b" ")?;
                write_inline(&mut p, " ")?;
            }

            HardBreak => {
                p.html.write_all(b"<br>")?;

                write_inline(&mut p, "\r\n")?;
                if p.state == State::Quote {
                    p.gmi.write_all(b"> ")?;
                }
            }

            Rule => {
                p.html.write_all(b"<hr>")?;

                write_paragraph_start(&mut p)?;
                p.gmi.write_all(b"---")?;
            }

            FootnoteReference(id) => {
                p.html.write_fmt(format_args!(
                    "<sup><a href=\"#ref-{}\">[{}]</a></sup>",
                    id, id,
                ))?;

                // [^1]
                p.gmi.write_all(b"[^")?;
                p.gmi.write_all(id.as_bytes())?;
                p.gmi.write_all(b"]")?;
            }

            TaskListMarker(done) => {
                if done {
                    p.html
                        .write_all(b"<input type=checkbox checked disabled>")?;
                    p.gmi.write_all(b"[x] ")?;
                } else {
                    p.html.write_all(b"<input type=checkbox disabled>")?;
                    p.gmi.write_all(b"[ ] ")?;
                }
            }

            other => {
                eprintln!("Unsupported event: {:?}", other);
            }
        }
    }

    // TODO: footnotes rendering

    p.html.flush()?;
    p.gmi.flush()?;

    Ok(())
}

#[inline(always)]
fn write_paragraph_start(p: &mut ParserVars) -> std::io::Result<()> {
    match p.state {
        State::Start | State::Quote => {}
        _ => {
            p.gmi.write_all(b"\r\n\r\n")?;
        }
    }
    Ok(())
}

#[inline(always)]
fn write_inline(p: &mut ParserVars, text: &str) -> std::io::Result<()> {
    match p.state {
        State::Link => {
            p.links.last_mut().unwrap().title.push_str(text);
        }
        State::Footnote => {
            p.notes.last_mut().unwrap().content.push_str(text);
        }
        _ => {}
    }
    p.gmi.write_all(text.as_bytes())?;
    Ok(())
}

struct ParserVars<'v> {
    html: Box<dyn std::io::Write>,
    gmi: Box<dyn std::io::Write>,

    state: State,
    ol: bool,
    counter: u64,

    links: Vec<GmiLink<'v>>,
    notes: Vec<Footnote<'v>>,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum State {
    Start,
    Paragraph,
    Link,
    Footnote,
    Quote,
    TableHead,
    Metadata {
        was_at_start: bool,
        kind: MetadataBlockKind,
    },
}

struct GmiLink<'link> {
    title: String,
    url: CowStr<'link>,
}

impl<'link> GmiLink<'link> {
    fn from_url(value: CowStr<'link>) -> Self {
        Self {
            title: String::new(),
            url: value,
        }
    }
}

struct Footnote<'note> {
    id: CowStr<'note>,
    content: String,
}

impl<'note> Footnote<'note> {
    fn from_id(value: CowStr<'note>) -> Self {
        Self {
            id: value,
            content: String::new(),
        }
    }
}
